import math
import re

from .input_handlers import validate_input

CALORIES_PER_G_FAT = 9
CALORIES_PER_G_CARB = 4
CALORIES_PER_G_PROTEIN = 4
CALORIES_PER_KG_BODY_FAT = 7700  # 3,500 kcal/lb
MAX_KG_LOSS_PER_WEEK = -0.907185  # -2 lbs
MAX_KG_GAIN_PER_WEEK = 0.907185  # 2 lbs
MAX_PERCENTAGE_BELOW_BMR = 0.2
GRAM_PROTEIN_LEAN_BODY_MASS_KG = 2.2
MAX_GRAM_PROTEIN_LEAN_BODY_MASS_KG = 3
GRAM_SATURATED_FAT_LEAN_BODY_MASS_KG = 0.15
MIN_FAT_G_LBM_KG = 0.75
MIN_FAT_SAFETY_FACTOR_PERCENTAGE = 0.1
OMEGA3_INTAKE_RANGE = (2, 6)  # Grams
LINOLEIC_ACID_AI_AGE_THRESHOLD = 50  # Years
LINOLEIC_ACID_AI = {
    'm': (17, 14),
    'f': (12, 11),
}
LBM_POP_DIST = (34, 72)  # 75-159 lbs
CHANGE_UNITS = ('%', 'kg', 'lbs')
VALID_CHANGE_UNITS = ('pct', '%', 'kg', 'lbs')
HEIGHT_UNITS = ('cm', 'in')
WEIGHT_UNITS = ('kg', 'lbs')


def lbs_to_kg(lbs: float) -> float:
    return lbs / 2.205


def kg_to_lbs(kgs: float) -> float:
    return kgs / 0.4536


def cm_to_inches(cms: float) -> float:
    return cms * 0.393701


def inches_to_cm(inches: float) -> float:
    return inches * 2.54


def cm_to_meters(cms: float) -> float:
    return cms / 100


def calculate_bmr(height, weight, age, sex):
    if sex == 'm':
        return int(88.362 + (13.397 * weight) + (4.799 * height) - (5.677 * age))
    elif sex == 'f':
        return int(447.593 + (9.247 * weight) + (3.098 * height) - (4.330 * age))


def calculate_bmi(height, weight):
    meters = cm_to_meters(height)
    return weight / meters ** 2


def bmi_to_bfp(bmi, sex, age):
    if sex == 'm':
        if age <= 15:
            return (1.51 * bmi) - (0.70 * age) - (3.6 * 1) + 1.4
        else:
            return (1.20 * bmi) + (0.23 * age) - (10.8 * 1) - 5.4
    elif sex == 'f':
        if age <= 15:
            return (1.51 * bmi) - (0.70 * age) - (3.6 * 0) + 1.4
        else:
            return (1.20 * bmi) + (0.23 * age) - (10.8 * 0) - 5.4


def calculate_fatty_acids(sex, age, lean_body_mass):
    ratio = (lean_body_mass - LBM_POP_DIST[0]) / (LBM_POP_DIST[1] - LBM_POP_DIST[0])
    omega3 = round(OMEGA3_INTAKE_RANGE[0] + (ratio * (OMEGA3_INTAKE_RANGE[1] - OMEGA3_INTAKE_RANGE[0])), 2)

    if omega3 < OMEGA3_INTAKE_RANGE[0]:
        omega3 = OMEGA3_INTAKE_RANGE[0]
    elif omega3 > OMEGA3_INTAKE_RANGE[1]:
        omega3 = OMEGA3_INTAKE_RANGE[1]

    if age < LINOLEIC_ACID_AI_AGE_THRESHOLD:
        linoleic_acid = LINOLEIC_ACID_AI[sex][0]
    else:
        linoleic_acid = LINOLEIC_ACID_AI[sex][1]

    saturated_fat = math.ceil(lean_body_mass * GRAM_SATURATED_FAT_LEAN_BODY_MASS_KG)

    total = math.ceil(lean_body_mass * MIN_FAT_G_LBM_KG * (1 + MIN_FAT_SAFETY_FACTOR_PERCENTAGE))
    total_essential = omega3 + linoleic_acid + saturated_fat

    if total < total_essential:
        total = total_essential

    return total, omega3, linoleic_acid, saturated_fat


def calculate_protein(lean_body_mass, extra_calories, split):
    remaining_calories = 0

    protein_grams_per_kg_to_reach_limit = MAX_GRAM_PROTEIN_LEAN_BODY_MASS_KG - GRAM_PROTEIN_LEAN_BODY_MASS_KG
    protein_grams_to_limit = protein_grams_per_kg_to_reach_limit * lean_body_mass
    protein_calories_to_limit = protein_grams_to_limit * CALORIES_PER_G_PROTEIN

    calories_per_macro = extra_calories / split
    if calories_per_macro > protein_calories_to_limit:
        protein_calories_to_add = protein_calories_to_limit
        remaining_calories = calories_per_macro - protein_calories_to_limit
    else:
        protein_calories_to_add = calories_per_macro

    return protein_calories_to_add, remaining_calories, calories_per_macro


def calculate_macros(tdee, distribution, change, change_units, weight, weight_units, body_fat_percentage, height,
                     height_units,
                     sex, age):

    validate_input(tdee, distribution, change, change_units, weight, weight_units, body_fat_percentage, height,
                   height_units,
                   sex, age)

    # Convert to kg
    if weight_units in ['lb', 'lbs', 'pounds']:
        weight = lbs_to_kg(weight)

    # Convert to cm
    if height_units in ['in', 'inches']:
        height = inches_to_cm(height)

    if body_fat_percentage:
        lean_body_mass = weight * (1 - body_fat_percentage / 100)
    else:
        bmi = calculate_bmi(height, weight)
        body_fat_percentage = bmi_to_bfp(bmi, sex, age)
        body_fat_percentage = round(body_fat_percentage, 2)
        lean_body_mass = weight * (1 - body_fat_percentage / 100)

    protein_grams = math.ceil(lean_body_mass * GRAM_PROTEIN_LEAN_BODY_MASS_KG)  # Set Minimum Protein
    fat_grams, omega3_grams, linoleic_acid_grams, saturated_fat_grams = calculate_fatty_acids(sex, age, lean_body_mass)

    protein_calories = protein_grams * CALORIES_PER_G_PROTEIN
    fat_calories = fat_grams * CALORIES_PER_G_FAT

    if change_units in ['pct', '%']:
        weight_change = weight * (change / 100)
    elif change_units in ['lbs', 'kg']:
        weight_change = change
    else:
        raise ValueError('Unrecognized change units')

    if weight_change < MAX_KG_LOSS_PER_WEEK:  # Max loss safety cap
        weight_change = MAX_KG_LOSS_PER_WEEK

    elif weight_change > MAX_KG_GAIN_PER_WEEK:  # Max gain safety cap
        weight_change = MAX_KG_GAIN_PER_WEEK

    daily_calorie_change = weight_change * CALORIES_PER_KG_BODY_FAT / 7
    daily_calories = int(tdee + daily_calorie_change)

    bmr = calculate_bmr(height, weight, age, sex)
    calories_lower_limit = bmr * (1 - MAX_PERCENTAGE_BELOW_BMR)

    if daily_calories < calories_lower_limit:
        daily_calories = calories_lower_limit

    warn = False
    if daily_calories < bmr:
        warn = True

    extra_calories = daily_calories - protein_calories - fat_calories
    if extra_calories < 0:
        extra_calories = 0

    carb_grams = 0
    if extra_calories:
        if distribution == 1:  # All Protein
            protein_grams += math.ceil(extra_calories / CALORIES_PER_G_PROTEIN)
        elif distribution == 2:  # All Carbs
            carb_grams = math.ceil(extra_calories / CALORIES_PER_G_CARB)
        elif distribution == 3:  # All Fat
            fat_grams += math.ceil(extra_calories / CALORIES_PER_G_FAT)
        elif distribution == 4:  # Mix Carbs & Fat
            carb_grams = math.ceil(extra_calories / 2 / CALORIES_PER_G_CARB)
            fat_grams += math.ceil(extra_calories / 2 / CALORIES_PER_G_FAT)
        elif distribution == 5:  # Mix Fat & Protein
            protein_calories_to_add, remaining_calories, calories_per_macro = calculate_protein(lean_body_mass,
                                                                                                extra_calories, 2)

            fat_grams += math.ceil((remaining_calories + calories_per_macro) / CALORIES_PER_G_FAT)
            protein_grams += math.ceil(protein_calories_to_add / CALORIES_PER_G_PROTEIN)
        elif distribution == 6:  # Mix Carbs & Protein
            protein_calories_to_add, remaining_calories, calories_per_macro = calculate_protein(lean_body_mass,
                                                                                                extra_calories, 2)

            carb_grams = math.ceil((remaining_calories + calories_per_macro) / CALORIES_PER_G_CARB)
            protein_grams += math.ceil(protein_calories_to_add / CALORIES_PER_G_PROTEIN)
        elif distribution == 7:  # Mix Carbs, Fat, & Protein
            protein_calories_to_add, remaining_calories, calories_per_macro = calculate_protein(lean_body_mass,
                                                                                                extra_calories, 3)

            additional_macro_calories = calories_per_macro + (remaining_calories / 2)

            carb_grams = math.ceil(additional_macro_calories / CALORIES_PER_G_CARB)
            fat_grams += math.ceil(additional_macro_calories / CALORIES_PER_G_FAT)
            protein_grams += math.ceil(protein_calories_to_add / CALORIES_PER_G_PROTEIN)

    protein_calories = int(protein_grams * CALORIES_PER_G_PROTEIN)
    fat_calories = int(fat_grams * CALORIES_PER_G_FAT)
    omega3_calories = int(omega3_grams * CALORIES_PER_G_FAT)
    linoleic_acid_calories = int(linoleic_acid_grams * CALORIES_PER_G_FAT)
    saturated_fat_calories = int(saturated_fat_grams * CALORIES_PER_G_FAT)
    carb_calories = int(carb_grams * CALORIES_PER_G_CARB)

    total_calories = protein_calories + fat_calories + carb_calories
    caloric_change = total_calories - tdee
    caloric_percentage_change = round((total_calories - tdee) / tdee * 100, 2)

    # Percentages
    protein_percentage = round(protein_calories / total_calories * 100, 2)
    fat_percentage = round(fat_calories / total_calories * 100, 2)
    omega3_percentage = round(omega3_calories / total_calories * 100, 2)
    linoleic_acid_percentage = round(linoleic_acid_calories / total_calories * 100, 2)
    saturated_fat_percentage = round(saturated_fat_calories / total_calories * 100, 2)
    carb_percentage = round(carb_calories / total_calories * 100, 2)

    # Lean Body Mass
    lean_body_mass_lbs = round(kg_to_lbs(lean_body_mass), 2)
    lean_body_mass = round(lean_body_mass, 2)

    # Grams / LBM
    protein_grams_kg_lbm = round(protein_grams / lean_body_mass, 2)
    protein_grams_lbs_lbm = round(protein_grams / lean_body_mass_lbs, 2)

    macros = {
        'protein': {
            'grams': {
                'total': protein_grams,
                'lean_body_mass': {
                    'kg': protein_grams_kg_lbm,
                    'lbs': protein_grams_lbs_lbm,
                }
            },
            'calories': protein_calories,
            'percentage': protein_percentage,
        },
        'fat': {
            'total': {
                'grams': fat_grams,
                'calories': fat_calories,
                'percentage': fat_percentage,
            },
            'omega3': {
                'grams': omega3_grams,
                'calories': omega3_calories,
                'percentage': omega3_percentage,
            },
            'linoleic_acid': {
                'grams': linoleic_acid_grams,
                'calories': linoleic_acid_calories,
                'percentage': linoleic_acid_percentage,
            },
            'saturated_fat': {
                'grams': saturated_fat_grams,
                'calories': saturated_fat_calories,
                'percentage': saturated_fat_percentage,
            }
        },
        'carb': {
            'grams': carb_grams,
            'calories': carb_calories,
            'percentage': carb_percentage,
        }
    }

    data = {
        'macros': macros,
        'stats': {
            'metabolism': {
                'total_calories': total_calories,
                'caloric_change': caloric_change,
                'caloric_percentage_change': caloric_percentage_change,
                'tdee': tdee,
                'basal_metabolic_rate': bmr,
            },
            'body_composition': {
                'lean_body_mass': {
                    'kg': lean_body_mass,
                    'lbs': lean_body_mass_lbs,
                },
                'body_fat_percentage': body_fat_percentage
            }
        },
        'warn': warn,
    }

    return data


def print_results(data):
    print('\n\n')
    print('===== YOUR DIETARY BREAKDOWN =====')
    print(
        f"1. Protein:\n   Amount: {data['macros']['protein']['grams']['total']}g\n   Calories from Protein: {data['macros']['protein']['calories']} kcal")
    print(f"   Protein Percentage: {data['macros']['protein']['percentage']:.2f}%")
    print(
        f"   Protein / Lean Body Mass: {data['macros']['protein']['grams']['lean_body_mass']['kg']:.2f} g/kg ({data['macros']['protein']['grams']['lean_body_mass']['lbs']:.2f} g/lb)\n")
    print(
        f"2. Fat: \n   Total Fat: {data['macros']['fat']['total']['grams']:.0f}g\n   Calories from Fat: {data['macros']['fat']['total']['calories']} kcal")
    print(f"   Fat Percentage: {data['macros']['fat']['total']['percentage']:.2f}%")
    print(
        f"   Omega-3 EPA & DHA:\n      Amount: {data['macros']['fat']['omega3']['grams']:.2f}g\n      Calories: {data['macros']['fat']['omega3']['calories']} kcal")
    print(
        f"   Linoleic Acid:\n      Amount: {data['macros']['fat']['linoleic_acid']['grams']}g\n      Calories: {data['macros']['fat']['linoleic_acid']['calories']} kcal")
    print(
        f"   Saturated Fat:\n      Amount: {data['macros']['fat']['saturated_fat']['grams']}g\n      Calories: {data['macros']['fat']['saturated_fat']['calories']} kcal\n")
    print(
        f"3. Carbs:\n   Amount: {data['macros']['carb']['grams']}g\n   Calories from Carbs: {data['macros']['carb']['calories']} kcal")
    print(f"   Carb Percentage: {data['macros']['carb']['percentage']:.2f}%")

    print(f"\nDaily Caloric Intake: {data['stats']['metabolism']['total_calories']:,} kcal")
    print(f"Daily Caloric Change: {data['stats']['metabolism']['caloric_change']:,} kcal")

    print(f"\nYour Body Stats:")
    print(
        f"Lean Body Mass: {data['stats']['body_composition']['lean_body_mass']['kg']:.0f} kg ({data['stats']['body_composition']['lean_body_mass']['lbs']:.0f} lbs)")
    print(f"Body Fat Percentage: {data['stats']['body_composition']['body_fat_percentage']:.2f}%")
    print(f"Total Daily Energy Expenditure: {data['stats']['metabolism']['tdee']:,} kcal")
    print(f"Basal Metabolic Rate: {data['stats']['metabolism']['basal_metabolic_rate']:,} kcal")

    if data['warn']:
        print()
        print('===== IMPORTANT MEDICAL WARNING =====')
        print(
            f'Your calculated daily calorie intake {data['stats']['metabolism']['total_calories']:,} kcal is below your estimated basal metabolic rate (BMR) of {data['stats']['metabolism']['basal_metabolic_rate']:,} kcal.')
        print(
            'Consuming fewer calories than your BMR for an extended period can lead to potential health risks, including nutrient deficiencies and metabolic slowdown.')
        print(
            'Ensure your diet includes a variety of nutrient-dense foods to meet your vitamin, mineral, and overall nutritional needs.')
        print(
            'Additionally, consider consulting with a healthcare provider or dietitian before starting any calorie-restricted diet, especially one below your BMR.')
        print(
            'They may suggest a tailored multivitamin or specific supplements based on your individual health profile and dietary needs.')


import re


CHANGE_UNITS = ('%', 'kg', 'lbs')
VALID_CHANGE_UNITS = ('pct', '%', 'kg', 'lbs')
HEIGHT_UNITS = ('cm', 'in')
WEIGHT_UNITS = ('kg', 'lbs')


def parse_input(user_input, available_units):
    user_input = user_input.lower()

    # Define the regex pattern to match numbers and units
    units = '|'.join(available_units)
    pattern = r"([-]?\d*\.?\d+)\s*(" + units + ")"

    # Search for matches in the user input
    match = re.search(pattern, user_input)

    if match:
        # Extract the number and unit from the matched groups
        number = float(match.group(1))  # Convert number to float
        unit = match.group(2)
        return number, unit
    else:
        raise ValueError()


def get_user_input():
    print("Press Ctrl-C to quit...")

    # TDEE
    while True:
        tdee = None
        try:
            print()
            tdee = int(input('What is your Total Daily Energy Expenditure? '))
        except ValueError:
            print()
            print('Error: Please enter a number.')
            continue
        if tdee < 0:
            print()
            print('Error: Please enter a positive number.')
            continue
        break

    # Weight
    while True:
        print()
        try:
            weight_input = input(f'What is your weight (units: {', '.join(WEIGHT_UNITS)})? ')
            weight, weight_units = parse_input(weight_input, available_units=WEIGHT_UNITS)
        except ValueError:
            print()
            print('Error: Please enter a number with lbs or kg as units.')
            continue
        if weight <= 0:
            print()
            print('Error: Please enter a positive number.')
            continue
        break

    while True:
        print()
        bfp_known = input('Do you know your Body Fat Percentage (Y/N)? ').lower()
        if bfp_known in ['y', 'n']:
            break
        else:
            print()
            print('Error: Please enter either Y or N.')
            continue

    if bfp_known == 'y':
        # Get BFP
        while True:
            print()
            try:
                body_fat_percentage = float(input('What is your body fat percentage? '))
            except ValueError:
                print()
                print('Error: Please enter a number.')
                continue
            if body_fat_percentage <= 0:
                print()
                print('Error: Please enter a positive number.')
                continue
            elif body_fat_percentage >= 100:
                print()
                print('Error: Body Fat Percentage must be between 0 and 100.')
                continue
            break
    else:
        body_fat_percentage = None

    # Height
    while True:
        print()
        try:
            height_input = input(f'What is your height (units: {', '.join(HEIGHT_UNITS)})? ')
            height, height_units = parse_input(height_input, available_units=HEIGHT_UNITS)
        except ValueError:
            print()
            print('Error: Please enter a number with inches or cm as units.')
            continue
        if height <= 0:
            print()
            print('Error: Please enter a positive number.')
            continue
        break

    # Sex
    while True:
        print()
        sex = input('What is your sex (M/F)? ').lower()
        if sex in ['m', 'f']:
            break
        else:
            print()
            print('Error: Please enter either M or F.')

    # Age
    while True:
        print()
        try:
            age = int(input('What is your age (years)? ').lower())
            break
        except ValueError:
            print()
            print('Error: Please enter a number.')

    while True:
        print()
        print(f'Desired change per week (units: {', '.join(CHANGE_UNITS)})?\n'
              'Example:\n'
              '-1.0% (Lose 1% per week)\n'
              '0.25 lbs (Gain 0.25 lbs per week)')
        try:
            change, change_units = parse_input(input('> '), available_units=CHANGE_UNITS)
        except ValueError:
            print()
            print(f'Error: Please enter a number with units ({', '.join(CHANGE_UNITS)}).')
            continue
        break

    while True:
        print()
        print('How would you like to distribute remaining calories?\n'
              '(1) Favor Protein\n'
              '(2) Favor Carbohydrates\n'
              '(3) Favor Fat\n'
              '(4) Blend of Carbs & Fat\n'
              '(5) Blend of Fat & Protein\n'
              '(6) Blend of Carbs & Protein\n'
              '(7) Blend of Carbs, Fat, & Protein')
        try:
            distribution = int(input('> '))
        except ValueError:
            print()
            print('Error: Please enter a number 1 through 7.')
            continue
        if not 1 <= distribution <= 7:
            print()
            print('Error: Please enter a number 1 through 7.')
            continue
        break

    return tdee, distribution, change, change_units, weight, weight_units, body_fat_percentage, height, height_units, sex, age


def validate_input(tdee, distribution, change, change_units, weight, weight_units, body_fat_percentage, height,
                   height_units,
                   sex, age):
    if type(tdee) == int:
        if tdee < 0:
            ValueError('Error: Please enter a positive number.')
    else:
        raise ValueError('Error: Please enter a positive number.')

    if type(distribution) == int:
        if not 1 <= distribution <= 7:
            raise ValueError('Error: Please enter a number between 1 and 7.')
    else:
        raise ValueError('Error: Please enter a number between 1 and 7.')

    if type(change) != float:
        raise ValueError('Error: Please enter a number.')

    if type(change_units) != str:
        raise ValueError('Error: Please enter units as strings.')
    elif change_units not in VALID_CHANGE_UNITS:
        raise ValueError(f'Error: Unknown change unit {change_units}, available units: {', '.join(VALID_CHANGE_UNITS)}')

    if type(weight) != float:
        raise ValueError('Error: Please enter a number.')

    if type(weight_units) != str:
        raise ValueError('Error: Please enter units as strings.')
    elif weight_units not in WEIGHT_UNITS:
        raise ValueError(f'Error: Unknown weight unit {weight_units}, available units: {', '.join(WEIGHT_UNITS)}')

    if body_fat_percentage:
        if type(body_fat_percentage) != float:
            raise ValueError('Error: Please enter a number.')
        if body_fat_percentage <= 0:
            raise ValueError('Error: Please enter a positive number.')
        if body_fat_percentage >= 100:
            raise ValueError('Error: Body Fat Percentage must be between 0 and 100.')

    if type(height) != float:
        raise ValueError('Error: Please enter a number.')
    elif height <= 0:
        raise ValueError('Error: Please enter a positive number.')

    if type(height_units) != str:
        raise ValueError('Error: Please enter units as strings.')
    elif height_units not in HEIGHT_UNITS:
        raise ValueError(f'Error: Unknown height unit {height_units}, available units: {', '.join(HEIGHT_UNITS)}')

    if type(sex) != str:
        raise ValueError('Error: Please enter a string')
    elif sex not in ['m', 'f']:
        raise ValueError('Error: Please enter either M or F.')

    if type(age) == float:
        age = int(age)
    if type(age) == int:
        if age <= 0:
            raise ValueError('Error: Please enter a positive number.')
    else:
        raise ValueError('Error: Please enter a number.')


# __main__.py
from advanced_macro_calculator.input_handlers import get_user_input
from advanced_macro_calculator.calculations import calculate_macros
from advanced_macro_calculator.outputs import print_results


def main():
    # Get user input
    user_input = get_user_input()

    # Perform calculations
    results = calculate_macros(*user_input)

    # Print or output the results
    print_results(results)


if __name__ == '__main__':
    main()
